#
# state.rb
#
#   Copyright (c) 1999,2000 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

require 'amstd/must'


module Racc

  class RaccError < StandardError; end


  #
  # ISet
  #
  # indexed set.
  # all items respond to :ident
  #

  class ISet

    def initialize( a = [] )
      @set = a
    end

    attr :set

    def add( i )
      @set[ i.ident ] = i
    end

    def []( key )
      @set[ key.ident ]
    end

    def []=( key, val )
      @set[ key.ident ] = val
    end

    alias include? []
    alias key? []

    def update( other )
      s = @set; o = other.set
      i = t = nil
      o.each_index {|i| if t = o[i] then s[i] = t end }
    end

    def update_a( a )
      s = @set
      i = nil
      a.each {|i| s[ i.ident ] = i }
    end

    def delete( key )
      i = @set[ key.ident ]
      @set[ key.ident ] = nil
      i
    end

    def each( &block )
      @set.compact.each( &block )
    end

    def to_a
      @set.compact
    end

    def to_s
      "[#{@set.compact.join(' ')}]"
    end
    alias inspect to_s

    def size
      @set.nitems
    end

    def empty?
      @set.nitems == 0
    end

    def clear
      @set.clear
    end

    def dup
      ISet.new @set.dup
    end
  
  end   # class ISet


  #######################################################################
  ###########################            ################################
  ###########################    rule    ################################
  ###########################            ################################
  #######################################################################


  #
  # RuleTable
  #
  # stands grammer. Each items of @rules are Rule object.
  #

  class RuleTable

    def initialize( racc )
      @racc        = racc
      @symboltable = racc.symboltable

      @verbose = racc.verbose
      @d_token = racc.d_token
      @d_rule  = racc.d_rule
      @d_state = racc.d_state

      @rules    = []
      @hashval  = 4   # size of dummy rule
      @start    = nil
      @tok_chk  = nil
    end


    def []( x )
      @rules[x]
    end

    def each_rule( &block )
      @rules.each( &block )
    end
    alias each each_rule

    def each_index( &block )
      @rules.each_index( &block )
    end

    def each_with_index( &block )
      @rules.each_with_index( &block )
    end

    def size
      @rules.size
    end

    def to_s
      "<Racc::RuleTable>"
    end


    def register( targ, slist, tempprec, act )
      rule = Rule.new(
        targ, slist, act,
        @rules.size + 1,         # ID
        @hashval,                # hash value
        tempprec                 # prec
      )
      @rules.push rule

      @hashval += slist.size + 2
    end

    def start=( sim )
      unless @start then
        @start = sim
        true
      else
        false
      end
    end
    
    attr :start

    def token_list=( list )
      if @tok_chk then
        bug! 'rule@tok_chk registered twice'
      end
      @tok_chk = list
    end


    def init
      $stderr.puts 'initializing values' if @verbose

      #
      # add dummy rule
      #
      tmp = Rule.new(
          @symboltable.dummy,
          [ @start, @symboltable.anchor, @symboltable.anchor ],
          UserAction.new( '', 0 ),
          0, 0, nil )
        # id hash prec
      @rules.unshift tmp
      @rules.freeze


      rule = ptr = orig = tmp = tok = t = s = nil

      #
      # t.heads
      #
      @rules.each do |rule|
        rule.target.heads.push rule.ptrs[0]
      end

      #
      # t.terminal?, self_null?
      #
      @symboltable.each do |t|
        t.term = (t.heads.size == 0)
        tmp = false
        t.heads.each do |ptr|
          if ptr.reduce? then
            tmp = true
            break
          end
        end
        t.snull = tmp
      end

      @symboltable.fix
      @symboltable.token_check @tok_chk if @tok_chk

      #
      # t.locate, rule.prec
      #
      @rules.each do |rule|
        tmp = nil
        rule.ptrs.each do |ptr|
          unless ptr.reduce? then
            tok = ptr.unref
            tok.locate.push ptr
            tmp = tok if tok.terminal?
          end
        end
        rule.prec ||= tmp
      end

      #
      # t.expand
      #
      @symboltable.each_nonterm {|t| compute_expand t }

      #
      # t.nullable?
      #
      @symboltable.each_nonterm do |t|
        tmp = false
        t.expand.each do |ptr|
          if tmp = ptr.reduce? then
            break
          end
          if tmp = ptr.unref.self_null? then
            break
          end
        end
        t.null = tmp
      end

    end


    def compute_expand( t )
      puts "expand> #{t.to_s}" if @d_token
      t.expand = coex( t, ISet.new, [] )
      puts "expand< #{t.to_s}: #{t.expand.to_s}" if @d_token
    end

    def coex( t, ret, lock )
      if tmp = t.expand then
        ret.update tmp
        return ret
      end

      tok = h = nil

      ret.update_a t.heads
      t.heads.each do |ptr|
        tok = ptr.deref
        if tok and tok.nonterminal? then
          unless lock[ tok.ident ] then
            lock[ tok.ident ] = true
            coex( tok, ret, lock )
          end
        end
      end

      ret
    end
    
  end   # class RuleTable


  #
  # Rule
  #
  # stands one rule of grammer.
  #

  class Rule

    def initialize( targ, syms, act, rid, hval, tprec )
      @target  = targ
      @symbols = syms
      @action  = act.val
      @lineno  = act.lineno
      @ident   = rid
      @hash    = hval
      @prec    = tprec
      @useless = false

      @ptrs = a = []
      syms.each_with_index do |t,i|
        a.push LocationPointer.new( self, i, t )
      end
      a.push LocationPointer.new( self, syms.size, nil )
    end

    attr :target
    attr :symbols

    attr :action
    attr :lineno

    attr :ident;  alias ruleid ident
    attr :hash
    attr :prec, true
    attr :ptrs

    def inspect
      "#<rule #{@ident} (#{@target})>"
    end

    def ==( other )
      Rule === other and @ident == other.ident
    end

    def []( idx )
      @symbols[idx]
    end

    def size
      @symbols.size
    end

    def to_s
      '#<rule#{@ident}>'
    end

    def accept?
      if tok = @symbols[-1] then
        tok.anchor?
      else
        false
      end
    end

    def useless=( f )
      if @useless then
        bug! 'rule.useless was set twice'
      end
      @useless = f
    end

    def useless?
      @useless
    end

    def each( &block )
      @symbols.each( &block )
    end

  end   # class Rule


  #
  # LocationPointer
  #
  # set of rule and position in it's rhs.
  # note that number of pointer is more than rule's rhs array,
  # because pointer points right of last symbol when reducing.
  #

  class LocationPointer

    def initialize( rl, idx, tok )
      @rule   = rl
      @index  = idx
      @deref  = tok

      @ident  = @rule.hash + @index
      @reduce = tok.nil?
    end

    attr :rule
    attr :index
    attr :deref
    alias unref deref

    attr :ident;      alias hash      ident
    attr :reduce;     alias reduce?   reduce

    def to_s
      sprintf( '(%d,%d %s)',
               @rule.ident, @index, reduce? ? '#' : unref.to_s )
    end
    alias inspect to_s

    def eql?( ot )
      @hash == ot.hash
    end
    alias == eql?

    def head?
      @index == 0
    end

    def increment
      @rule.ptrs[ @index + 1 ] or ptr_bug!
    end

    def decrement
      @rule.ptrs[ @index - 1 ] or ptr_bug!
    end

    def before( len )
      @rule.ptrs[ @index - len ] or ptr_bug!
    end

    private
    
    def ptr_bug!
      bug! "pointer not exist: self: #{to_s}"
    end

  end   # class LocationPointer


  #########################################################################
  ###########################              ################################
  ###########################    symbol    ################################
  ###########################              ################################
  #########################################################################

  #
  # SymbolTable
  #
  # the table of symbols.
  # each items of @symbols are Sym
  #

  class SymbolTable

    include Enumerable

    def initialize( racc )
      @chk     = {}
      @symbols = []
      
      @dummy  = get( :$start )
      @anchor = get( :$end )    # ID 0
      @error  = get( :error )   # ID 1

      @anchor.conv = 'false'
      @error.conv = 'Object.new'
    end

    attr :dummy
    attr :anchor
    attr :error

    def fix
      #
      # initialize table
      #

      term = []
      nt = []
      t = i = nil
      @symbols.each do |t|
        (t.terminal? ? term : nt).push t
      end
      @symbols = term + nt
      @nt_base = term.size

      @symbols.each_with_index do |t, i|
        t.ident = i
      end
    end

    def token_check( list )
      #
      # check if decleared symbols are really terminal
      #

      toks = @symbols[ 2, @nt_base - 2 ]
      list.uniq!
      list.each do |t|
        unless toks.delete t then
          $stderr.puts "racc warning: terminal #{t} decleared but not used"
        end
      end
      toks.each do |t|
        $stderr.puts "racc warning: terminal #{t} used but not decleared"
      end
    end


    def get( val )
      unless ret = @chk[ val ] then
        @chk[ val ] = ret = Sym.new( val )
        @symbols.push ret
      end

      ret
    end

    def []( id )
      @symbols[id]
    end

    attr :nt_base

    def nt_max
      @symbols.size
    end

    def each( &block )
      @symbols.each &block
    end

    def each_terminal( &block )
      @symbols[ 0, @nt_base ].each( &block )
    end

    def each_nonterm( &block )
      @symbols[ @nt_base, @symbols.size - @nt_base ].each( &block )
    end

  end


  #
  # Sym
  #
  # stands symbol (terminal and nonterminal).
  # This class is not named Symbol because there is
  # a class 'Symbol' after ruby 1.5.
  #

  class Sym

    def initialize( val )
      @ident = nil
      @value = val

      @term  = nil
      @nterm = nil
      @conv  = nil
      @prec  = nil

      @heads    = []
      @locate   = []
      @snull    = nil
      @null     = nil
      @expand   = nil

      @useless  = nil

      # for human
      @to_s =
        if @value.respond_to? 'id2name' then @value.id2name
                                        else @value.to_s.inspect
        end
      # for ruby source
      @uneval =
        if @value.respond_to? 'id2name' then ':' + @value.id2name
                                        else @value.to_s.inspect
        end
    end


    class << self
      def once_writer( nm )
        nm = nm.id2name
        module_eval %-
          def #{nm}=( v )
            bug! unless @#{nm}.nil?
            @#{nm} = v
          end
        -
      end
    end

    #
    # attrs
    #

    once_writer :ident
    attr :ident
    alias hash ident

    attr :value

    def terminal?()    @term end
    def nonterminal?() @nterm end

    def term=( t )
      bug! unless @term.nil?
      @term = t
      @nterm = !t
    end

    def conv=( str ) @conv = @uneval = str end
    attr :conv

    attr_accessor :prec
    attr_accessor :assoc

    def to_s()   @to_s.dup end
    def uneval() @uneval.dup end
    alias inspect to_s

    #
    # computed
    #

    attr :heads
    attr :locate

    once_writer :snull
    def self_null?() @snull end

    once_writer :null
    def nullable?() @null end

    once_writer :expand
    attr :expand

    once_writer :useless
    def useless?() @useless end

  end   # class Sym



  ########################################################################
  ###########################             ################################
  ###########################    state    ################################
  ###########################             ################################
  ########################################################################


  #
  # StateTable
  #
  # the table of lalr states.
  #

  class StateTable

    def initialize( racc )
      @ruletable   = racc.ruletable
      @symboltable = racc.symboltable
      @anchor_t    = racc.symboltable.anchor

      @verbose = racc.verbose
      @prof    = racc.make_profile
      @d_state = racc.d_state
      @d_la    = racc.d_la
      @d_prec  = racc.d_prec

      @states = []
      @statecache = {}

      @actions = ActionTable.new( @ruletable, self )
    end

    attr :actions

    def size
      @states.size
    end

    def inspect
      '#<state table>'
    end

    alias to_s inspect

    def []( i )
      @states[i]
    end

    def each_state( &block )
      @states.each( &block )
    end

    alias each each_state

    def each_index( &block )
      @states.each_index( &block )
    end


    ###
    ### nfa
    ###

    def init
      $stderr.puts 'generating states' if @verbose

      # add state 0
      core_to_state( [ @ruletable[0].ptrs[0] ] )

      cur = 0
      @gotos = []
      while cur < @states.size do
        generate_states @states[cur]   # state is added here
        cur += 1
      end

      @actions.init
    end

    def generate_states( state )
      puts "dstate: #{state}" if @d_state

      table = {}
      ptr = pt = s = g = nil

      state.closure.each do |ptr|
        if sym = ptr.deref then
          addsym table, sym, ptr.increment
        end
      end

      table.each do |sym, core|
        puts "dstate: sym=#{sym} ncore=#{core}" if @d_state

        dest = core_to_state( core.to_a )
        state.goto_table[sym] = dest
        id = if sym.nonterminal? then @gotos.size else nil end
        g = Goto.new( id, sym, state, dest )
        if sym.nonterminal? then
          @gotos.push g
        end
        state.gotos[sym] = g
        puts "dstate: #{state.ident} --#{sym}-> #{dest.ident}" if @d_state

        # check infinite recursion
        if state.ident == dest.ident and state.closure.size == 1 then
          raise RaccError,
            sprintf( "Infinite recursion: state %d, with rule %d",
                     state.ident, state.ptrs[0].rule.ident )
        end
      end
    end

    def addsym( table, sym, ptr )
      unless s = table[sym] then
        table[sym] = s = ISet.new
      end
      s.add ptr
    end

    def core_to_state( core )
      #
      # creates/find state keeping core unique
      #

      k = fingerprint( core )
      unless dest = @statecache[k] then
        # not registered yet
        dest = State.new( @states.size, core )
        @states.push dest

        @statecache[k] = dest
        
        puts "core_to_state: create state   ID #{dest.ident}" if @d_state
      else
        if @d_state then
          puts "core_to_state: dest is cached ID #{dest.ident}"
          puts "core_to_state: dest core #{dest.core.join(' ')}"
        end
      end

      dest
    end

    def fingerprint( arr )
      arr.collect {|i| i.ident }.pack( 'L*' )
    end


    ###
    ### dfa
    ###

    def resolve
      #
      # lookahead algorithm ver.3 -- bison 1.26
      #

      $stderr.puts "resolving #{@states.size} states" if @verbose

      state = goto = st = rl = a = t = g = nil
      b = Time.times.utime


      gotos = @gotos
      if @d_la then
        puts "\n--- goto ---"
        gotos.each_with_index {|g,i| print i, ' '; p g }
      end

      ### initialize_LA()
      ### set_goto_map()
      la_rules = []
      @states.each do |state|
        state.check_la( la_rules )
      end


      ### initialize_F()
      f     = Array.new( gotos.size, 0 )   # use Integer as bitmap
      reads = []
      edge  = []
      gotos.each do |goto|
        goto.to_state.goto_table.each do |t, st|
          if t.terminal? then
            f[goto.ident] |= (1 << t.ident)
          elsif t.nullable? then
            edge.push goto.to_state.gotos[t].ident
          end
        end
        if edge.empty? then
          reads.push nil
        else
          reads.push edge
          edge = []
        end
      end
      digraph f, reads
      if @d_la then
        puts "\n--- F1 (reads) ---"
        print_tab gotos, reads, f
      end


      ### build_relations()
      ### compute_FOLLOWS
      path = nil
      edge = []
      lookback = Array.new( la_rules.size, nil )
      includes = []
      gotos.each do |goto|
        goto.symbol.heads.each do |ptr|
          path = record_path( goto.from_state, ptr.rule )
          g = path[-1]
          st = g ? g.to_state : goto.from_state
          if st.conflict? then
            addrel lookback, st.rruleid(ptr.rule), goto
          end
          path.reverse_each do |g|
            break if     g.symbol.terminal?
            edge.push    g.ident
            break unless g.symbol.nullable?
          end
        end
        if edge.empty? then
          includes.push nil
        else
          includes.push edge
          edge = []
        end
      end
      includes = transpose( includes )
      digraph f, includes
      if @d_la then
        puts "\n--- F2 (includes) ---"
        print_tab gotos, includes, f
      end


      ### compute_lookaheads
      la = Array.new( la_rules.size, 0 )
      lookback.each_with_index do |arr, i|
        if arr then
          arr.each do |g|
            la[i] |= f[g.ident]
          end
        end
      end
      if @d_la then
        puts "\n--- LA (lookback) ---"
        print_tab la_rules, lookback, la
      end

      #
      # set LA tokens
      #
      @states.each do |state|
        state.set_la la
        determine state
      end


      if @verbose then
        e = Time.times.utime
        $stderr.puts "all resolved in #{e - b} sec"
      end

      set_accept
      pack_states
      check_useless
    end

    def addrel( tbl, i, item )
      if a = tbl[i] then
        a.push item
      else
        tbl[i] = [item]
      end
    end

    def record_path( begst, rule )
      st = begst
      path = []
      rule.symbols.each do |t|
        goto = st.gotos[t]
        path.push goto
        st = goto.to_state
      end
      path
    end

    def transpose( rel )
      new = Array.new( rel.size, nil )
      rel.each_with_index do |arr, idx|
        if arr then
          arr.each do |i|
            addrel new, i, idx
          end
        end
      end
      new
    end

    def digraph( map, relation )
      n = relation.size
      index    = Array.new( n, nil )
      vertices = []
      @infinity = n + 2

      i = nil
      index.each_index do |i|
        if not index[i] and relation[i] then
          traverse i, index, vertices, map, relation
        end
      end
    end

    def traverse( i, index, vertices, map, relation )
      vertices.push i
      index[i] = height = vertices.size

      proci = nil
      if rp = relation[i] then
        rp.each do |proci|
          unless index[proci] then
            traverse proci, index, vertices, map, relation
          end
          if index[i] > index[proci] then
            # circulative recursion !!!
            index[i] = index[proci]
          end
          map[i] |= map[proci]
        end
      end

      if index[i] == height then
        while true do
          proci = vertices.pop
          index[proci] = @infinity
          break if i == proci

          map[proci] |= map[i]
        end
      end
    end

    ###

    def print_atab( idx, tab )
      tab.each_with_index do |i,ii|
        printf '%-20s', idx[ii].inspect
        p i
      end
    end

    def print_tab( idx, rel, tab )
      tab.each_with_index do |bin,i|
        print i, ' ', idx[i].inspect, ' << '; p rel[i]
        print '  '
        each_t( @symboltable, bin ) {|t| print ' ', t }
        puts
      end
    end

    def printb( i )
      each_t( @symboltable, i ) do |t|
        print t, ' '
      end
      puts
    end

    def each_t( tbl, set )
      i = ii = idx = nil
      0.upto( set.size ) do |i|
        (0..7).each do |ii|
          if set[ idx = i * 8 + ii ] == 1 then
            yield tbl[idx]
          end
        end
      end
    end

    ###

    def determine( state )
      if state.conflict? then
        resolve_rr state, state.ritems
        resolve_sr state, state.shift
      else
        if state.reduce.empty? then
          # shift
          state.shift.each do |t|
            state.action[t] = @actions.shift( state.goto_table[t] )
          end
        else
          # reduce
          state.defact = @actions.reduce( state.reduce[0] )
        end
      end
    end

    def resolve_rr( state, r )
      pt = t = act = item = nil

      r.each do |item|
        item.each_la( @symboltable ) do |t|
          act = state.action[t]
          if act then
            Reduce === act or bug! "#{act.type} in action table"
            #
            # can't resolve R/R conflict (on t).
            #   reduce with upper rule as default
            #
            state.rr_conflict act.rule, item.rule, t
          else
            # not conflict
            state.action[t] = @actions.reduce( item.rule )
          end
        end
      end
    end

    def resolve_sr( state, s )
      stok = rtok = goto = act = nil

      s.each do |stok|
        goto = state.goto_table[stok]
        act = state.action[stok]

        unless act then
          # no conflict
          state.action[ stok ] = @actions.shift( goto )
        else
          unless Reduce === act then
p stok
p act
state.action.each do |k,v|
  print k.inspect, ' ', v.inspect, "\n"
end
            bug! "#{act.type} in action table"
          end

          # conflict on stok

          rtok = act.rule.prec
          case do_resolve_sr( stok, rtok )
          when :Reduce        # action is already set

          when :Shift         # overwrite
            act.decref
            state.action[ stok ] = @actions.shift( goto )

          when :Remove        # remove
            act.decref
            state.action.delete stok

          when :CantResolve   # shift as default
            act.decref
            state.action[ stok ] = @actions.shift( goto )
            state.sr_conflict stok, act.rule
          end
        end
      end
    end
    
   def do_resolve_sr( stok, rtok )
      puts "resolve_sr: s/r conflict: rtok=#{rtok}, stok=#{stok}" if @d_prec

      unless rtok and rtok.prec then
        puts "resolve_sr: no prec for #{rtok}(R)" if @d_prec
        return :CantResolve
      end
      rprec = rtok.prec

      unless stok and stok.prec then
        puts "resolve_sr: no prec for #{stok}(S)" if @d_prec
        return :CantResolve
      end
      sprec = stok.prec

      ret = if rprec == sprec then
              case rtok.assoc
              when :Left     then :Reduce
              when :Right    then :Shift
              when :Nonassoc then :Remove  # 'rtok' makes error
              else
                bug! "#{rtok}.assoc is not Left/Right/Nonassoc"
              end
            else
              if rprec > sprec
              then :Reduce
              else :Shift
              end
            end

      puts "resolve_sr: resolved as #{ret.id2name}" if @d_prec
      ret
    end


    def set_accept
      anch = @anchor_t
      init_state = @states[0].goto_table[ @ruletable.start ]
      targ_state = init_state.action[ anch ].goto_state
      acc_state  = targ_state.action[ anch ].goto_state

      acc_state.action.clear
      acc_state.goto_table.clear
      acc_state.defact = @actions.accept
    end

    def pack_states
      st = arr = act = nil
      i = n = s = r = nil

      @states.each do |st|
        #
        # find most frequently used reduce rule
        #
        act = st.action
        arr = Array.new( @ruletable.size, 0 )
        act.each do |t,a|
          if Reduce === a then
            arr[ a.ruleid ] += 1
          end
        end
        i = s = nil
        i = arr.max
        s = arr.index( i ) if i > 0

        if s then
          r = @actions.reduce(s)
          if not st.defact or st.defact == r then
            act.delete_if {|t,a| a == r }
            st.defact = r
          end
        else
          st.defact ||= @actions.error
        end
      end
    end

    def check_useless
      act = nil
      used = []
      @actions.each_reduce do |act|
        if not act or act.refn == 0 then
          act.rule.useless = true
        else
          t = act.rule.target
          used[ t.ident ] = t
        end
      end
      @symboltable.nt_base.upto( @symboltable.nt_max - 1 ) do |n|
        unless used[n] then
          @symboltable[n].useless = true
        end
      end
    end

  end   # class StateTable


  #
  # State
  #
  # stands lalr state.
  #

  class State

    def initialize( ident, core )
      @ident = ident
      @core  = core
      @items = nil

      @goto_table = {}
      @gotos      = {}

      @action = {}
      @defact = nil

      @resolve_log = []

      @rrconf = nil
      @srconf = nil

      ###

      @closure = make_closure( @core )
    end


    attr :ident
    alias stateid ident
    alias hash ident

    attr :core
    attr :closure

    attr :goto_table
    attr :gotos

    attr :shift
    attr :reduce
    attr :ritems

    attr :action
    attr :defact, true   # default action

    attr :resolve_log

    attr :rrconf
    attr :srconf

    def inspect
      "<state #{@ident}>"
    end

    alias to_s inspect

    def ==( oth )
      @ident == oth.ident
    end

    alias eql? ==


    def make_closure( core )
      set = ISet.new
      core.each do |ptr|
        set.add ptr
        if t = ptr.deref and t.nonterminal? then
          set.update_a t.expand
        end
      end
      set.to_a
    end

    def check_la( la_rules )
      @conflict = false
      s = []
      r = []
      @closure.each do |ptr|
        if t = ptr.deref then
          if t.terminal? then
            s[t.ident] = t
            if t.id == 1 then   # $error
              @conflict = true
            end
          end
        else
          r.push ptr.rule
        end
      end
      unless r.empty? then
        if not s.empty? or r.size > 1 then
          @conflict = true
        end
      end
      s.compact!
      @shift  = s
      @reduce = r

      if @conflict then
        @la_rules_i = la_rules.size
        @la_rules = r.collect {|i| i.ident }
        la_rules.concat r
      else
        @la_rules_i = @la_rules = nil
      end
    end

    def conflict?
      @conflict
    end

    def rruleid( rule )
      if i = @la_rules.index( rule.ident ) then
        @la_rules_i + i
      else
        puts '/// rruleid'
        p self
        p rule
        p @reduce
        p @la_rules_i
        bug! 'rruleid'
      end
    end

    def set_la( la )
      return unless @conflict

      i = @la_rules_i
      @ritems = r = []
      @reduce.each do |rule|
        r.push Item.new( rule, la[i] )
        i += 1
      end
    end

    def rr_conflict( high, low, ctok )
      c = RRconflict.new( @ident, high, low, ctok )

      unless @rrconf then
        @rrconf = {}
      end
      if a = @rrconf[ctok] then
        a.push c
      else
        @rrconf[ctok] = [c]
      end
    end

    def sr_conflict( shift, reduce )
      c = SRconflict.new( @ident, shift, reduce )

      unless @srconf then
        @srconf = {}
      end
      if a = @srconf[shift] then
        a.push c
      else
        @srconf[shift] = [c]
      end
    end

  end   # State


  #
  # Goto
  #
  # stands one transition on the grammer.
  # REAL GOTO means transition by nonterminal,
  # but this class treats also terminal's.
  # In case of terminal transition, .ident is nil.
  #

  class Goto

    def initialize( ident, sym, from, to )
      @ident      = ident
      @symbol     = sym
      @from_state = from
      @to_state   = to
    end

    attr :ident
    attr :symbol
    attr :from_state
    attr :to_state
    
    def inspect
      "(#{@from_state.ident}-#{@symbol}->#{@to_state.ident})"
    end
  
  end


  #
  # Item
  #
  # lalr item. set of rule and its lookahead tokens.
  #

  class Item
  
    def initialize( rule, la )
      @rule = rule
      @la  = la
    end

    attr :rule
    attr :la

    def each_la( tbl )
      la = @la
      i = ii = idx = nil
      0.upto( la.size ) do |i|
        (0..7).each do |ii|
          if la[ idx = i * 8 + ii ] == 1 then
            yield tbl[idx]
          end
        end
      end
    end

  end


  #
  # ActionTable
  #
  # the table of lalr actions. Actions are
  # Shift, Reduce, Accept and Error
  #

  class ActionTable

    def initialize( rt, st )
      @ruletable = rt
      @statetable = st

      @reduce = []
      @shift = []
      @accept = nil
      @error = nil
    end

    def init
      @ruletable.each do |rl|
        @reduce.push Reduce.new( rl )
      end
      @statetable.each do |st|
        @shift.push Shift.new( st )
      end
      @accept = Accept.new
      @error = Error.new
    end


    def reduce_n
      @reduce.size
    end

    def reduce( i )
      if Rule === i then
        i = i.ident
      else
        i.must Integer
      end

      unless ret = @reduce[i] then
        bug! "reduce action #{i} not exist"
      end
      ret.incref
      ret
    end

    def each_reduce( &block )
      @reduce.each &block
    end


    def shift_n
      @shift.size
    end

    def shift( i )
      if State === i then
        i = i.ident
      else
        i.must Integer
      end

      @shift[i] or bug! "shift action #{i} not exist"
    end

    def each_shift( &block )
      @shift.each &block
    end


    attr :accept
    attr :error

  end


  class Shift

    def initialize( goto )
      @goto_state = goto
    end

    attr :goto_state

    def goto_id
      @goto_state.ident
    end

    def inspect
      "<shift #{@goto_state.ident}>"
    end

  end


  class Reduce

    def initialize( rule )
      @rule = rule
      @refn = 0
    end

    attr :rule
    attr :refn

    def ruleid
      @rule.ident
    end

    def inspect
      "<reduce #{@rule.ident}>"
    end

    def incref
      @refn += 1
    end

    def decref
      @refn -= 1
      if @refn < 0 then
        bug! 'act.refn < 0'
      end
    end

  end


  class Accept

    def inspect
      "<accept>"
    end

  end


  class Error

    def inspect
      "<error>"
    end
  
  end


  #
  # Conflicts
  #

  class SRconflict

    def initialize( sid, shift, reduce )
      @stateid = sid
      @shift   = shift
      @reduce  = reduce
    end
    
    attr :stateid
    attr :shift
    attr :reduce
  
    def to_s
      sprintf( 'state %d: S/R conflict rule %d reduce and shift %s',
               @stateid, @reduce.ruleid, @shift.to_s )
    end

  end

  class RRconflict
    
    def initialize( sid, high, low, tok )
      @stateid   = sid
      @high_prec = high
      @low_prec  = low
      @token     = tok
    end

    attr :stateid
    attr :high_prec
    attr :low_prec
    attr :token
  
    def to_s
      sprintf( 'state %d: R/R conflict with rule %d and %d on %s',
               @stateid, @high_prec.ident, @low_prec.ident, @token.to_s )
    end

  end

end
